import { PromiseWithCancel } from 'real-cancellable-promise';
import { Codec } from '@adam-rocska/ts-codec';

declare class TaskError<I, O> extends Error {
    readonly taskName: string;
    readonly input?: I;
    readonly error?: any;
    readonly output?: O;
    constructor(taskName: string, message: string, input?: I, error?: any, output?: O);
}

interface TaskOutput<I, O> {
    readonly input: I;
    readonly output: O;
}
declare function isTaskOutput<I, O>(value: object): value is TaskOutput<I, O>;

interface TaskProcess<I, O> {
    readonly input: I;
    task: Promise<O[]> | PromiseWithCancel<O[]>;
}
declare function isTaskProcess<I, O>(value: object): value is TaskProcess<I, O>;

interface TaskQueueHook<I, O> {
    readonly input: readonly I[];
    readonly process: readonly TaskProcess<I, O>[];
    readonly output: readonly TaskOutput<I, O>[];
    readonly error: readonly TaskError<I, O>[];
    push(input: I | I[]): void;
    kill(process: 'all'): void;
    kill(process: TaskProcess<I, O>): void;
    flush(target: 'output', ...values: TaskOutput<I, O>[]): TaskOutput<I, O>[];
    flush(target: 'error', ...values: TaskError<I, O>[]): TaskError<I, O>[];
}
declare function isTaskQueueHook<I, O>(value: object): value is TaskQueueHook<I, O>;

interface Task<I, O> {
    (input: I): O[] | PromiseWithCancel<O[]> | Promise<O[]>;
}

interface TaskQueueDescriptor<I, O> {
    readonly name: string;
    readonly codec: Codec<I>;
    readonly task: Task<I, O>;
    readonly input?: TaskQueueHook<any, I>;
    readonly precondition?: (input: I) => boolean;
    readonly postcondition?: (input: I, output: O) => boolean;
}
declare const TaskQueueDescriptor: <I, O>(descriptor: TaskQueueDescriptor<I, O>) => TaskQueueDescriptor<I, O>;
declare const isTaskQueueDescriptor: (candidate: any) => candidate is TaskQueueDescriptor<unknown, unknown>;

declare function useTaskQueue<I, O>(descriptor: TaskQueueDescriptor<I, O>): TaskQueueHook<I, O>;

declare class InconsistencyError extends Error {
    readonly knownDescriptor: TaskQueueDescriptor<any, any>;
    readonly mismatchingDescriptor: TaskQueueDescriptor<any, any>;
    constructor(knownDescriptor: TaskQueueDescriptor<any, any>, mismatchingDescriptor: TaskQueueDescriptor<any, any>);
}

declare function nullTaskQueue<I, O>(): TaskQueueHook<I, O>;
declare function isNullTaskQueue<I, O>(queue: TaskQueueHook<I, O>): boolean;

export { InconsistencyError, type Task, TaskError, type TaskOutput, type TaskProcess, TaskQueueDescriptor, type TaskQueueHook, isNullTaskQueue, isTaskOutput, isTaskProcess, isTaskQueueDescriptor, isTaskQueueHook, nullTaskQueue, useTaskQueue };
